# -*- coding: utf8 -*-

from django.db import models
from django.utils.translation import gettext_lazy as _

import string
from datetime import datetime

# require Tags Field http://softwaremaniacs.org/soft/tags/en/
from djengel.tags.models import Tag
from djengel.tags import fields

class ExpenseItemSummary(dict) :
    """集計結果用のクラス。費目モデルを再利用しても構わなかったが、辞書を継承して作成してみた。

    object.propertyで辞書のキー、値にアクセス可能

    費目に設定してあるひと月のリミット金額に対して出費がある程度(80%と100%)を超えているかどうかのメソッドを持っている。
    """
    def __init__(self, pay_date=None, name=None, amount=0, expense_id=0, max_amount=0) :
        self['pay_date']    = pay_date
        self['name']        = name
        self['amount']     = amount
        self['expense_id']  = expense_id
        self['max_amount'] = max_amount

    def __getattr__(self, name):
        try: return self[name]
        except: object.__getattribute__(self, name)

    def is_eighty(self) :
        return self._percentage(80)

    def is_over(self) :
        return self._percentage(100)
    
    def _percentage(self, border=100) :
        if self.max_amount == 0 or self.amount == 0 :
            return False
        elif int(float(self.amount) / float(self.max_amount) * 100) >= border :
            return True
        return False


class ExpenseItemManager(models.Manager) :
    """費目モデルのマネージャ。デフォルトのQuerysetに、非表示は取得しない、という条件を付与している。
    """
    def get_query_set(self) :
        return super(ExpenseItemManager, self).get_query_set().filter(visible=True)

class ExpenseItemSummaryManager(models.Manager) :
    """費目ごとのサマリー用マネージャ

    家計簿程度のものであれば、group byを使用するまでもなく、ORMでモデルをぐるぐる回して計算してもたいしたことはない(Djangoはcursorのfetchも一気にやる訳ではない)が、サンプル的に直SQLを放った。PostgreSQLで動作する。円グラフ用(ひと月分)も折れ線グラフ用(指定期間分)も同じSQL分を利用している。
    """

    def get_summary_by_year_month(self, date_from='1970/01', date_to='2099/12') :
        df = datetime(int(date_from.split('/')[0]),int(date_from.split('/')[1]), 1)
        y_e = int(date_to.split('/')[0])
        m_e = int(date_to.split('/')[1])
        if m_e == 12 :
            y_e += 1
            m_e  = 0
        dt = datetime(y_e, m_e + 1, 1)
        return self._get_summary('yyyy/mm', df.strftime('%Y/%m/%d'), dt.strftime('%Y/%m/%d'))

    def get_summary_by_month(self, year=1970, month=1) :
        d = datetime(int(year), int(month), 1)
        return self.get_summary_by_year_month(d.strftime('%Y/%m/%d'), d.strftime('%Y/%m/%d'))

    def get_summary_month_to_date(self) :
        dn = datetime.now()
        return self.get_summary_by_month(dn.year, dn.month)

    _sum_sql = """
            SELECT to_char(o.pay_date, %s) as pay_date, ei.name as name, sum(o.amount) as amount, ei.id  as ei_id, ei.max_amount as max_amount
             FROM squander_outgoing o, squander_expenseitem ei
             WHERE 
              o.expense_item_id = ei.id AND
              o.pay_date >= to_date(%s, 'yyyy/mm/dd') AND
              o.pay_date < to_date(%s, 'yyyy/mm/dd') 
             GROUP BY to_char(o.pay_date, %s), name, ei_id, max_amount
             ORDER BY pay_date, ei_id"""
        
    def _get_summary(self, format='yyyy/mm/dd', date_from=None, date_to=None) :
        from django.db import connection
        cursor = connection.cursor()
        cursor.execute(self._sum_sql, [format, date_from, date_to, format])
        result_list = []
        for row in cursor.fetchall() :
            summ = ExpenseItemSummary(row[0], row[1], row[2], row[3], row[4])
            result_list.append(summ)
        return result_list

class ExpenseItem(models.Model) :
    """費目を表すモデル。

    デフォルト動作の変更をいくつか実装してある。

        - デフォルトのdeleteを上書き
        - デフォルトのマネージャの上書き
        - カスタムマネージャの追加。
    """
    name = models.CharField(_('Expense Item Name'), blank=False, maxlength=30)
    max_amount = models.IntegerField(_('Max Amount'), blank=False, default=0, help_text=_('Max amount per month.'))
    food_cost = models.BooleanField(_('Food Cost Flag'), default=False, help_text=_('weather food cost or not'))
    visible = models.BooleanField(_('Visible Flag'), default=True, help_text=_('if expense item is invisible, you can\'t edit anymore.'))
    objects = ExpenseItemManager() #デフォルトのマネージャをカスタムマネージャに変更
    summary_objects = ExpenseItemSummaryManager() #サマリー用のマネージャを登録

    class Admin :
        list_display = ("name", "max_amount", "visible", "food_cost")
        ordering = ['id']

    class Meta :
        verbose_name = _('Expense Item')
        verbose_name_plural = _('Expense Items')

    def __str__(self) :
        """オブジェクトの人間可読表現
        """
        return self.name

    def delete(self) :
        """費目を物理削除すると関連する出費も含めて削除されてしまうため、モデルを利用しての削除はできないようにしてある。

        代わりに、削除を呼ぶと「表示フラグ」をオフにするようにし、同時にマネージャをカスタムマネージャとし、デフォルトのQuerysetに「表示フラグ」オンという条件を入れるようにした。これにより、費目を管理画面から削除すると、「削除した費目を管理画面で編集できない」「削除した費目は出費の追加画面に候補として出てこない」という動作を実現している。
        """
        self.visible = False
        self.save()

class Outgoing(models.Model) :
    """出費を表すモデル。

    複数のタグを関連づけることができる。タグはロシア人のIvan Sagalaevさん作によるTagアプリケーションをdjengelに取り込んでいます(djengel/tags)。

        - http://softwaremaniacs.org/soft/tags/en/

    ただし、タグを正規化された形でのみ保持しているので、管理画面や明細画面で恐ろしい回数のSQL発行を行ってしまっています(当然出費を取得して、その出費に関連するタグを取得して、ロジックでまわせば2度のSQL発行ですみます)。

    本当はこれはしたくないので、普段は非正規化している。TSVにしたTagのキャッシュ情報を持たせるということです(Djangoの場合はTagモデルのsaveやdeleteメソッドを上書きすれば管理が簡単にできる)。
    """
    pay_date = models.DateField(_('Pay Date'), help_text=_('payment date'))
    amount = models.IntegerField(_('Amount'), blank=True, null=True, core=True, help_text=_('payment amount'))
    expense_item = models.ForeignKey(ExpenseItem, verbose_name=_('Expense Item'))
    #enable TAGGING
    tags = fields.TagsField(Tag, blank=True, help_text=_('Tags'))
    note = models.TextField(_('Note'), blank=True, help_text=_('write anything you like.'))

    class Meta :
        verbose_name = _('Outgoing') #このモデルの単数系名(かつ国際化)
        verbose_name_plural = _('Outputgoings') #このモデルの複数形名(かつ国際化)
        ordering = ['-pay_date'] #管理画面でのデフォルト並び順は支払日の降順
        permissions = (
            ("can_view", "Can View"),
        ) #Adminを宣言すると自動で追加される「追加」「変更」「削除」と別のカスタム権限を宣言している

    class Admin :
        date_hierarchy = "pay_date" #管理画面でどの日付を利用して絞り込むことができるかの指定
        list_display = ("pay_date", "amount", "short_note", "tag") #管理画面の一覧に出す項目。プロパティを利用するとフィールド以外も出力可能。
        list_filter = ('pay_date', 'expense_item', 'tags') #管理画面のフィルターにどの項目を利用可能とするか。日付、ブーリアン、外部キーが利用可能
        search_fields = ['note'] #管理画面の文字検索で、どのフィールドを検索対象にするか。複数指定すると、orで検索される。
        save_as = True #既存データの編集画面に、新しいレコードとして追加するためのボタンを表示する
        js = (
           'js/tags.js',
        ) #管理画面に追加でリンクするJavaScriptの指定

    def short_note(self) :
        """管理画面に出力するためにメモの1行目だけを返す
        """
        return self.note.split('\n')[0]
    short_note.short_description = _('Note Summary')

    def _get_tag_lists(self) :
        return self.tags.all() 
    tag_list = property(_get_tag_lists)

    def _get_joint_tag(self) :
        """出費の一覧でタグを表示するために、タグをスペース区切りで単一の文字列にして返す。

        joinだとうまく出力されない。 __str__ がjoinでうまく機能しない？ 一時期動いてたんだけど・・
        """
        str = ''
        for t in self._get_tag_lists() :
            str += ' %s' % t.value
        return str
    tag = property(_get_joint_tag)

    def __str__(self) :
        """オブジェクトの人間可読表現
        """
        return '%s(%d円) %s' % (self.pay_date,self.amount,self.short_note())
